package NGCP::InfluxDB::HTTP;

use strict;
use warnings;

use Carp;
use HTTP::Tiny;
use JSON::XS;

our $VERSION = '0.90';

sub new
{
    my ($this, %opts) = @_;
    my $class = ref $this || $this;

    my $self = {
        addr => undef,
    };

    my %http_attr = (
        agent => "InfluxDB-HTTP/$VERSION",
        timeout => $opts{timeout} // 180,
    );
    my $http;
    my $host = $opts{host} // 'localhost';
    my $port = $opts{port} // 8086;
    $host = lc $host;
    $http = HTTP::Tiny->new(%http_attr);
    $self->{addr} = "http://$host:$port";
    $self->{http} = $http;

    bless $self, $class;

    return $self;
}

sub get_http
{
    my $self = shift;

    return $self->{http};
}

sub ping
{
    my $self = shift;
    my $uri = $self->_get_influxdb_http_api_uri('ping');
    my $response = $self->{http}->head($uri);

    if (! $response->{success}) {
        return $self->_get_influxdb_response_error($response);
    }

    my $version = $response->{headers}{'x-influxdb-version'};
    return {
        retval => 1,
        version => $version,
    };
}

sub query
{
    my ($self, $query, %opts) = @_;

    if (not defined $query) {
        croak 'missing query argument';
    }
    if (ref $query eq 'ARRAY') {
        $query = join ';', @$query;
    }
    if (defined $opts{epoch}) {
        croak 'invalid epoch argument' if $opts{epoch} !~ m/^(h|m|s|ms|u|ns)$/;
    }

    my $uri_query = { q => $query };
    $uri_query->{db} = $opts{database} if defined $opts{database};
    $uri_query->{chunk_size} = $opts{chunk_size} if defined $opts{chunk_size};
    $uri_query->{epoch} = $opts{epoch} if defined $opts{epoch};
    my $uri = $self->_get_influxdb_http_api_uri('query', $uri_query);

    my $response = $self->{http}->post($uri);

    if (! $response->{success}) {
        return $self->_get_influxdb_response_error($response);
    }

    my $data = decode_json($response->{content});

    return {
        retval      => 1,
        data        => $data,
        results     => $data->{results},
        request_id  => $response->{headers}{'request-id'},
        json        => $response->{content},
    };
}


sub write ## no critic (Subroutines::ProhibitBuiltinHomonyms)
{
    my ($self, $measurement, %opts) = @_;

    if (not defined $measurement) {
        croak 'missing measurement argument';
    }
    if (ref $measurement eq 'ARRAY') {
        $measurement = join("\n", @$measurement);
    }

    my $uri_query;
    $uri_query->{db} = $opts{database} if defined $opts{database};
    $uri_query->{consistency} = $opts{consistency} if defined $opts{consistency};
    $uri_query->{precision} = $opts{precision} if defined $opts{precision};
    $uri_query->{rp} = $opts{rp} if defined $opts{rp};
    my $uri = $self->_get_influxdb_http_api_uri('write', $uri_query);

    my $response = $self->{http}->post($uri, { content => $measurement });

    if ($response->{status} != 204) {
        return $self->_get_influxdb_response_error($response);
    }

    return {
        retval => 1,
    };
}

sub _get_influxdb_http_api_uri
{
    my ($self, $endpoint, $query_data) = @_;

    my $uri = "$self->{addr}/$endpoint";
    if (defined $query_data) {
        my $uri_params = $self->{http}->www_form_urlencode($query_data);
        $uri .= "?$uri_params";
    }

    return $uri;
}

sub _get_influxdb_response_error
{
    my ($self, $response) = @_;

    my $response_error;
    eval {
        my $data = decode_json($response->{content});
        $response_error = $data->{error};
    } or do {
        $response_error = "$response->{status} $response->{reason}";
    };

    return {
        retval => 0,
        error => $response_error,
        http_code => $response->{status},
        http_error => $response->{reason},
    };
}

1;

__END__

=head1 NAME

NGCP::InfluxDB::HTTP - a way to interact with InfluxDB

=head1 VERSION

Version 0.01

=head1 DESCRIPTION

NGCP::InfluxDB::HTTP allows you top interact with the InfluxDB HTTP API. The
module essentially provides one method per InfluxDB HTTP API endpoint, that is
C<ping>, C<write> and C<query>.

    use NGCP::InfluxDB::HTTP;

    my $influx = NGCP::InfluxDB::HTTP->new();

    my $ping_result = $influx->ping();
    print "$ping_result->{retval}\n";

    my $query = $influx->query(
        [ 'SELECT Lookups FROM _internal.monitor.runtime WHERE time > '.(time - 60)*1000000000, 'SHOW DATABASES'],
        epoch => 's',
    );

    print "$query->{json}\n";


=head1 SUBROUTINES/METHODS

=head2 RETURN VALUES AND ERROR HANDLING

The results from subroutines are returned as a HASHREF. The key C<retval>
stores a boolean denoting whether the sub failed or succeeded. On error
the C<error> key stores the error message.

    my $ping = $influx->ping();
    print $ping->{version} unless $ping->{retval};

=head2 new host => 'localhost', port => 8086

Passing C<host> and/or C<port> is optional, defaulting to the InfluxDB
defaults.

Returns an instance of NGCP::InfluxDB::HTTP.

=head2 ping

Pings the InfluxDB instance configured in the constructor (i.e. by C<host>
and C<port>).

Returned HASHREF contains a C<retval> key which evaluates to true or false
depending on whether the ping was successful or not. If true, then it
contains a C<version> key that indicates the InfluxDB version running on
the pinged server.

The C<version> key is extracted from the C<X-Influxdb-Version> HTTP
response header, which is part of the HTTP response from the pinged
InfluxDB instance.

    my $ping = $influx->ping();
    print $ping->{version} if $ping->{retval};

=head2 query query, database => "DATABASE", chunk_size => CHUNK_SIZE, epoch => "ns"

Used to query the InfluxDB instance. All parameters but the first one are
optional. The C<query> parameter can either be a String or a Perl ArrayRef
of Strings, where every String contains a valid InfluxDB query.

If the returned HASHREF C<retval> key evaluates to true, indicating that the
query was successful, then the returned object's C<data> key contains the
entire response from InfluxDB as Perl hash. Additionally the C<request_id>
key provides the request identifier as set in the HTTP response headers by
InfluxDB. This can for example be useful for correlating requests with log
files. The C<json> key contains the raw JSON response.

=head2 write measurement, database => "DATABASE", consistency => "CONSISTENCY", precision => "PRECISION", rp => "RETENTION_POLICY"

Writes data into InfluxDB. The parameter C<measurement> can either be a
String or an ArrayRef of Strings, where each String contains one valid
InfluxDB LineProtocol statement. All of those mesaurements are then sent
to InfluxDB and the specified database.

The returned HASHREF C<retval> key evaluates to true if the write was
successful, and otherwise to false. In which case the C<error> key contains
the error message.

=head2 get_http

Returns the internally used HTTP::Tiny instance for possible modifications
(e.g. to configure an HTTP proxy).

=head1 AUTHOR

Raphael Seebacher, C<< <raphael@seebachers.ch> >>
Guillem Jover, C<< <gjover@sipwise.com> >>

=head1 BUGS

Please report any bugs to Sipwise GmbH. The upstream code can be found at
<https://github.com/raphaelthomas/InfluxDB-HTTP>.

=head1 LICENSE

MIT License

Copyright (c) 2016 Raphael Seebacher
Copyright (c) 2016-2018 Sipwise GmbH, Austria.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

=cut
